var fs = require('fs');
var path = require("path");
var consoleFileLogger = require("../sl-logger.js").consoleFileLogger;
var sourceMap = require("source-map");
var globalErrorHandler = require('../global-error-handler');
var sourceRootResolver = require("./source-root-resolver");
var fileAndFolderUtils = require('../file-and-folder-utils');
var utils = require('../utils');
var SlJsInfra = require("sl-js-infra").SlJsInfra;
var SourceMapReader = SlJsInfra.SourceMapReader;


// TODO: refactor to normal module
function readSourceMaps(fullFilename, fileContent, logger) {
    var log = logger ? logger.child({
        methodName: "readSourceMaps"
    }) : consoleFileLogger("readSourceMaps");
    if (!fullFilename || !fileContent){
        log.warn("'fullFilename' and 'fileContent' cannot be null or empty");
        log.debug("fullFilename: '%s'", fullFilename);
        return null;
    };

    try {
        var reader = new SourceMapReader(fullFilename, logger);
        var sourceMapData = reader.readSourceMapData(fileContent);
        if (!sourceMapData) {
            return null;
        }
        var consumer = new sourceMap.SourceMapConsumer(sourceMapData);
        tryOverridePathInSourceMap(consumer, fullFilename, logger);
        return consumer;
    }catch (e) {
        logger.warn("Failed while reading source map for '%s'", fullFilename);
        logger.warn(e);
        return null;
    }
}

/**
 * By default, source maps store relative path to the file.
 * This methods try to convert the path to an absolute path for easier processing
 * later in the build process.
 */
function tryOverridePathInSourceMap(consumer, fullFilename, logger){
    if (!consumer || !consumer.sources || consumer.sources.length == 0){
        return;
    }

    var log = logger ? logger.child({
        methodName: "tryOverridePathInSourceMap"
    }) : consoleFileLogger("tryOverridePathInSourceMap");

    var parsedPath = path.parse(fullFilename);
    var newSources = [];

    for (var i=0; i < consumer.sources.length; i++){
        var sourceFile =  consumer["_sources"]["_array"][i];
        var absolutePath = "";
        if (path.isAbsolute(sourceFile)){
            absolutePath = sourceFile;
            if (!sourceRootResolver.isResolved) {
                log.debug("Source map path's are absolute, considering source root as resolved.");
                sourceRootResolver.isResolved = true;
            }
        }else {
            var sourceRoot = sourceRootResolver.resolveSourceRoot(sourceFile, parsedPath.dir, log);
            sourceFile = fileAndFolderUtils.removeRelativeSigns(sourceFile);
            absolutePath = path.join(sourceRoot, sourceFile);
        }

        log.debug("Path '" + sourceFile + "' is overridden to '" + absolutePath+ "'.");
        newSources.push(absolutePath);
    }

    consumer["_sources"]["_array"] = newSources;

    if (consumer.sources.length > 1){
        log.warn("File '" + fullFilename+ "' has multiple sources. An assumption is made that they all originate can be found in folder '" + parsedPath.dir + "'.");
    }
}

function resolveRelativeName(fullFilename, sourceRoot) {
    var fileName = fullFilename.replace(sourceRoot, "");
    fileName = utils.adjustPathSlashes(fileName);
    if (fileName.startsWith("/")) { //TODO: What happens when we are on Linux and that's not relative.
        fileName = fileName.substring(1);
    }
    return fileName;
}

function resolveFileName(relativeFileName, fullFilename, log) {
    var fileDir = path.dirname(fullFilename);
    //This code assumes that the fileDir is the same file where the Source Maps are located.
    //TODO: Understand why we have this assumption?
    sourceRootResolver.resolveSourceRoot(fullFilename, fileDir, log);
    var sourceRoot = sourceRootResolver.sourceRoot;
    if (sourceRoot) {
        relativeFileName = resolveRelativeName(fullFilename, sourceRoot);
    }


    return relativeFileName;
}

function ActualFile(relativeFileName, fullFilename) {
    var content;
    relativeFileName = utils.adjustPathSlashes(relativeFileName);
    this.getFilename = function ActualFile$getFilename(logger) {
        return resolveFileName(relativeFileName, fullFilename, logger);
    }
    this.getFullFilename = function ActualFile$getFullFilename() {
        return fullFilename; //TODO: Think if this one should be with fwd slashes as well.
    }
    this.getContent = function ActualFile$getContent() {
        if(!content) {
            content = fs.readFileSync(fullFilename, 'utf-8');
        }
        return content;
    }
    this.readSourceMaps = function ActualFile$readSourceMaps(logger) {
        return readSourceMaps(fullFilename, this.getContent(), logger);
    }
}
module.exports = ActualFile;
